"""Example of sourced answer search.

The Linkup search can also be used to perform direct Question Answering, with `output_type` set to
`sourcedAnswer`. In this case, the API will output an answer to the query in natural language,
along with the sources supporting it.

To use this script, copy the `.env.example` file at the root of the repository inside a `.env`, and
fill the missing values, or pass a Linkup API key to the `LinkupClient` initialization.
"""

from dotenv import load_dotenv
from rich import print

from linkup import LinkupClient

load_dotenv()
client = LinkupClient()

response = client.search(
    query="What are the 3 major events in the life of Abraham Lincoln ?",
    depth="standard",  # or "deep"
    output_type="sourcedAnswer",
    include_inline_citations=False,
)
print(response)



import os

import pytest

from linkup import LinkupClient


@pytest.fixture(scope="session")
def client() -> LinkupClient:
    if os.getenv("LINKUP_API_KEY") is None:
        os.environ["LINKUP_API_KEY"] = "<linkup-api-key>"
    return LinkupClient()



import json
import os
from datetime import date
from typing import Any, Literal, Optional, Union

import httpx
from pydantic import BaseModel

from linkup._version import __version__
from linkup.errors import (
    LinkupAuthenticationError,
    LinkupFailedFetchError,
    LinkupInsufficientCreditError,
    LinkupInvalidRequestError,
    LinkupNoResultError,
    LinkupTooManyRequestsError,
    LinkupUnknownError,
)
from linkup.types import (
    LinkupFetchResponse,
    LinkupSearchResults,
    LinkupSearchStructuredResponse,
    LinkupSourcedAnswer,
)


class LinkupClient:
    """The Linkup Client class, providing functions to call the Linkup API endpoints using Python.

    Args:
        api_key: The API key for the Linkup API. If None, the API key will be read from the
            environment variable `LINKUP_API_KEY`.
        base_url: The base URL for the Linkup API, for development purposes.

    Raises:
        ValueError: If the API key is not provided and not found in the environment variable.
    """

    __version__ = __version__

    def __init__(
        self,
        api_key: Optional[str] = None,
        base_url: str = "https://api.linkup.so/v1",
    ) -> None:
        if api_key is None:
            api_key = os.getenv("LINKUP_API_KEY")
        if not api_key:
            raise ValueError("The Linkup API key was not provided")

        self.__api_key = api_key
        self.__base_url = base_url

    def search(
        self,
        query: str,
        depth: Literal["standard", "deep"],
        output_type: Literal["searchResults", "sourcedAnswer", "structured"],
        structured_output_schema: Union[type[BaseModel], str, None] = None,
        include_images: Optional[bool] = None,
        from_date: Optional[date] = None,
        to_date: Optional[date] = None,
        exclude_domains: Optional[list[str]] = None,
        include_domains: Optional[list[str]] = None,
        include_inline_citations: Optional[bool] = None,
        include_sources: Optional[bool] = None,
    ) -> Any:
        """Perform a web search using the Linkup API `search` endpoint.

        All optional parameters will default to the Linkup API defaults when not provided. The
        Linkup API defaults are available in the
        [official documentation](https://docs.linkup.so/pages/documentation/api-reference/endpoint/post-search).

        Args:
            query: The search query.
            depth: The depth of the search. Can be either "standard", for a straighforward and
                fast search, or "deep" for a more powerful agentic workflow.
            output_type: The type of output which is expected: "searchResults" will output raw
                search results, "sourcedAnswer" will output the answer to the query and sources
                supporting it, and "structured" will base the output on the format provided in
                structured_output_schema.
            structured_output_schema: If output_type is "structured", specify the schema of the
                output. Supported formats are a pydantic.BaseModel or a string representing a
                valid object JSON schema.
            include_images: Indicate whether images should be included during the search.
            from_date: The date from which the search results should be considered. If None, the
                search results will not be filtered by date.
            to_date: The date until which the search results should be considered. If None, the
                search results will not be filtered by date.
            exclude_domains: If you want to exclude specific domains from your search.
            include_domains: If you want the search to only return results from certain domains.
            include_inline_citations: If output_type is "sourcedAnswer", indicate whether the
                answer should include inline citations.
            include_sources: If output_type is "structured", indicate whether the answer should
                include sources. This will modify the schema of the structured response.

        Returns:
            The Linkup API search result, which can have different types based on the parameters:
            - LinkupSearchResults if output_type is "searchResults"
            - LinkupSourcedAnswer if output_type is "sourcedAnswer"
            - the provided pydantic.BaseModel or an arbitrary data structure if output_type is
              "structured" and include_sources is False
            - LinkupSearchStructuredResponse with the provided pydantic.BaseModel or an arbitrary
              data structure as data field, if output_type is "structured" and include_sources is
              True

        Raises:
            TypeError: If structured_output_schema is not provided or is not a string or a
                pydantic.BaseModel when output_type is "structured".
            LinkupInvalidRequestError: If structured_output_schema doesn't represent a valid object
                JSON schema when output_type is "structured".
            LinkupAuthenticationError: If the Linkup API key is invalid.
            LinkupInsufficientCreditError: If you have run out of credit.
            LinkupNoResultError: If the search query did not yield any result.
        """
        params: dict[str, Union[str, bool, list[str]]] = self._get_search_params(
            query=query,
            depth=depth,
            output_type=output_type,
            structured_output_schema=structured_output_schema,
            include_images=include_images,
            from_date=from_date,
            to_date=to_date,
            exclude_domains=exclude_domains,
            include_domains=include_domains,
            include_inline_citations=include_inline_citations,
            include_sources=include_sources,
        )

        response: httpx.Response = self._request(
            method="POST",
            url="/search",
            json=params,
            timeout=None,
        )
        if response.status_code != 200:
            self._raise_linkup_error(response=response)

        return self._parse_search_response(
            response=response,
            output_type=output_type,
            structured_output_schema=structured_output_schema,
            include_sources=include_sources,
        )

    async def async_search(
        self,
        query: str,
        depth: Literal["standard", "deep"],
        output_type: Literal["searchResults", "sourcedAnswer", "structured"],
        structured_output_schema: Union[type[BaseModel], str, None] = None,
        include_images: Optional[bool] = None,
        from_date: Optional[date] = None,
        to_date: Optional[date] = None,
        exclude_domains: Optional[list[str]] = None,
        include_domains: Optional[list[str]] = None,
        include_inline_citations: Optional[bool] = None,
        include_sources: Optional[bool] = None,
    ) -> Any:
        """Asynchronously perform a web search using the Linkup API `search` endpoint.

        All optional parameters will default to the Linkup API defaults when not provided. The
        Linkup API defaults are available in the
        [official documentation](https://docs.linkup.so/pages/documentation/api-reference/endpoint/post-search).

        Args:
            query: The search query.
            depth: The depth of the search. Can be either "standard", for a straighforward and
                fast search, or "deep" for a more powerful agentic workflow.
            output_type: The type of output which is expected: "searchResults" will output raw
                search results, "sourcedAnswer" will output the answer to the query and sources
                supporting it, and "structured" will base the output on the format provided in
                structured_output_schema.
            structured_output_schema: If output_type is "structured", specify the schema of the
                output. Supported formats are a pydantic.BaseModel or a string representing a
                valid object JSON schema.
            include_images: Indicate whether images should be included during the search.
            from_date: The date from which the search results should be considered. If None, the
                search results will not be filtered by date.
            to_date: The date until which the search results should be considered. If None, the
                search results will not be filtered by date.
            exclude_domains: If you want to exclude specific domains from your search.
            include_domains: If you want the search to only return results from certain domains.
            include_inline_citations: If output_type is "sourcedAnswer", indicate whether the
                answer should include inline citations.
            include_sources: If output_type is "structured", indicate whether the answer should
                include sources. This will modify the schema of the structured response.

        Returns:
            The Linkup API search result, which can have different types based on the parameters:
            - LinkupSearchResults if output_type is "searchResults"
            - LinkupSourcedAnswer if output_type is "sourcedAnswer"
            - the provided pydantic.BaseModel or an arbitrary data structure if output_type is
              "structured" and include_sources is False
            - LinkupSearchStructuredResponse with the provided pydantic.BaseModel or an arbitrary
              data structure as data field, if output_type is "structured" and include_sources is
              True

        Raises:
            TypeError: If structured_output_schema is not provided or is not a string or a
                pydantic.BaseModel when output_type is "structured".
            LinkupInvalidRequestError: If structured_output_schema doesn't represent a valid object
                JSON schema when output_type is "structured".
            LinkupAuthenticationError: If the Linkup API key is invalid.
            LinkupInsufficientCreditError: If you have run out of credit.
            LinkupNoResultError: If the search query did not yield any result.
        """
        params: dict[str, Union[str, bool, list[str]]] = self._get_search_params(
            query=query,
            depth=depth,
            output_type=output_type,
            structured_output_schema=structured_output_schema,
            include_images=include_images,
            from_date=from_date,
            to_date=to_date,
            exclude_domains=exclude_domains,
            include_domains=include_domains,
            include_inline_citations=include_inline_citations,
            include_sources=include_sources,
        )

        response: httpx.Response = await self._async_request(
            method="POST",
            url="/search",
            json=params,
            timeout=None,
        )
        if response.status_code != 200:
            self._raise_linkup_error(response=response)

        return self._parse_search_response(
            response=response,
            output_type=output_type,
            structured_output_schema=structured_output_schema,
            include_sources=include_sources,
        )

    def fetch(
        self,
        url: str,
        include_raw_html: Optional[bool] = None,
        render_js: Optional[bool] = None,
        extract_images: Optional[bool] = None,
    ) -> LinkupFetchResponse:
        """Fetch the content of a web page using the Linkup API `fetch` endpoint.

        All optional parameters will default to the Linkup API defaults when not provided. The
        Linkup API defaults are available in the
        [official documentation](https://docs.linkup.so/pages/documentation/api-reference/endpoint/post-fetch).

        Args:
            url: The URL of the web page to fetch.
            include_raw_html: Whether to include the raw HTML of the webpage in the response.
            render_js: Whether the API should render the JavaScript of the webpage.
            extract_images: Whether the API should extract images from the webpage and return them
                in the response.

        Returns:
            The response of the web page fetch, containing the web page content.

        Raises:
            LinkupInvalidRequestError: If the provided URL is not valid.
            LinkupFailedFetchError: If the provided URL is not found or can't be fetched.
        """
        params: dict[str, Union[str, bool]] = self._get_fetch_params(
            url=url,
            include_raw_html=include_raw_html,
            render_js=render_js,
            extract_images=extract_images,
        )

        response: httpx.Response = self._request(
            method="POST",
            url="/fetch",
            json=params,
            timeout=None,
        )
        if response.status_code != 200:
            self._raise_linkup_error(response=response)

        return self._parse_fetch_response(response=response)

    async def async_fetch(
        self,
        url: str,
        include_raw_html: Optional[bool] = None,
        render_js: Optional[bool] = None,
        extract_images: Optional[bool] = None,
    ) -> LinkupFetchResponse:
        """Asynchronously fetch the content of a web page using the Linkup API `fetch` endpoint.

        All optional parameters will default to the Linkup API defaults when not provided. The
        Linkup API defaults are available in the
        [official documentation](https://docs.linkup.so/pages/documentation/api-reference/endpoint/post-fetch).

        Args:
            url: The URL of the web page to fetch.
            include_raw_html: Whether to include the raw HTML of the webpage in the response.
            render_js: Whether the API should render the JavaScript of the webpage.
            extract_images: Whether the API should extract images from the webpage and return them
                in the response.

        Returns:
            The response of the web page fetch, containing the web page content.

        Raises:
            LinkupInvalidRequestError: If the provided URL is not valid.
            LinkupFailedFetchError: If the provided URL is not found or can't be fetched.
        """
        params: dict[str, Union[str, bool]] = self._get_fetch_params(
            url=url,
            include_raw_html=include_raw_html,
            render_js=render_js,
            extract_images=extract_images,
        )

        response: httpx.Response = await self._async_request(
            method="POST",
            url="/fetch",
            json=params,
            timeout=None,
        )
        if response.status_code != 200:
            self._raise_linkup_error(response=response)

        return self._parse_fetch_response(response=response)

    def _user_agent(self) -> str:  # pragma: no cover
        return f"Linkup-Python/{self.__version__}"

    def _headers(self) -> dict[str, str]:  # pragma: no cover
        return {
            "Authorization": f"Bearer {self.__api_key}",
            "User-Agent": self._user_agent(),
        }

    def _request(
        self,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> httpx.Response:  # pragma: no cover
        with httpx.Client(base_url=self.__base_url, headers=self._headers()) as client:
            return client.request(
                method=method,
                url=url,
                **kwargs,
            )

    async def _async_request(
        self,
        method: str,
        url: str,
        **kwargs: Any,
    ) -> httpx.Response:  # pragma: no cover
        async with httpx.AsyncClient(base_url=self.__base_url, headers=self._headers()) as client:
            return await client.request(
                method=method,
                url=url,
                **kwargs,
            )

    def _raise_linkup_error(self, response: httpx.Response) -> None:
        error_data = response.json()

        if "error" in error_data:
            error = error_data["error"]
            code = error.get("code", "")
            error_msg = error.get("message", "")
            details = error.get("details", [])

            if details and isinstance(details, list):
                for detail in details:
                    if isinstance(detail, dict):
                        field = detail.get("field", "")
                        field_message = detail.get("message", "")
                        error_msg += f" {field}: {field_message}"

            if response.status_code == 400:
                if code == "SEARCH_QUERY_NO_RESULT":
                    raise LinkupNoResultError(
                        "The Linkup API returned a no result error (400). "
                        "Try rephrasing you query.\n"
                        f"Original error message: {error_msg}."
                    )
                if code == "FETCH_ERROR":
                    raise LinkupFailedFetchError(
                        "The Linkup API returned a fetch error (400). "
                        "The provided URL might not be found or can't be fetched.\n"
                        f"Original error message: {error_msg}."
                    )
                else:
                    raise LinkupInvalidRequestError(
                        "The Linkup API returned an invalid request error (400). Make sure the "
                        "parameters you used are valid (correct values, types, mandatory "
                        "parameters, etc.) and you are using the latest version of the Python "
                        "SDK.\n"
                        f"Original error message: {error_msg}."
                    )
            elif response.status_code == 401:
                raise LinkupAuthenticationError(
                    "The Linkup API returned an authentication error (401). Make sure your API "
                    "key is valid.\n"
                    f"Original error message: {error_msg}."
                )
            elif response.status_code == 403:
                raise LinkupAuthenticationError(
                    "The Linkup API returned an authorization error (403). Make sure your API "
                    "key is valid.\n"
                    f"Original error message: {error_msg}."
                )
            elif response.status_code == 429:
                if code == "INSUFFICIENT_FUNDS_CREDITS":
                    raise LinkupInsufficientCreditError(
                        "The Linkup API returned an insufficient credit error (429). Make sure "
                        "you haven't exhausted your credits.\n"
                        f"Original error message: {error_msg}."
                    )
                elif code == "TOO_MANY_REQUESTS":
                    raise LinkupTooManyRequestsError(
                        "The Linkup API returned a too many requests error (429). Make sure "
                        "you not sending too many requests at a time.\n"
                        f"Original error message: {error_msg}."
                    )
                else:
                    raise LinkupUnknownError(
                        "The Linkup API returned an invalid request error (429). Make sure the "
                        "parameters you used are valid (correct values, types, mandatory "
                        "parameters, etc.) and you are using the latest version of the Python "
                        "SDK.\n"
                        f"Original error message: {error_msg}."
                    )
            else:
                raise LinkupUnknownError(
                    f"The Linkup API returned an unknown error ({response.status_code}).\n"
                    f"Original error message: ({error_msg})."
                )

    def _get_search_params(
        self,
        query: str,
        depth: Literal["standard", "deep"],
        output_type: Literal["searchResults", "sourcedAnswer", "structured"],
        structured_output_schema: Union[type[BaseModel], str, None],
        include_images: Optional[bool],
        from_date: Optional[date],
        to_date: Optional[date],
        exclude_domains: Optional[list[str]],
        include_domains: Optional[list[str]],
        include_inline_citations: Optional[bool],
        include_sources: Optional[bool],
    ) -> dict[str, Union[str, bool, list[str]]]:
        params: dict[str, Union[str, bool, list[str]]] = dict(
            q=query,
            depth=depth,
            outputType=output_type,
        )

        if structured_output_schema is not None:
            if isinstance(structured_output_schema, str):
                params["structuredOutputSchema"] = structured_output_schema
            elif issubclass(structured_output_schema, BaseModel):
                json_schema: dict[str, Any] = structured_output_schema.model_json_schema()
                params["structuredOutputSchema"] = json.dumps(json_schema)
            else:
                raise TypeError(
                    f"Unexpected structured_output_schema type: '{type(structured_output_schema)}'"
                )
        if include_images is not None:
            params["includeImages"] = include_images
        if from_date is not None:
            params["fromDate"] = from_date.isoformat()
        if to_date is not None:
            params["toDate"] = to_date.isoformat()
        if exclude_domains is not None:
            params["excludeDomains"] = exclude_domains
        if include_domains is not None:
            params["includeDomains"] = include_domains
        if include_inline_citations is not None:
            params["includeInlineCitations"] = include_inline_citations
        if include_sources is not None:
            params["includeSources"] = include_sources

        return params

    def _get_fetch_params(
        self,
        url: str,
        include_raw_html: Optional[bool],
        render_js: Optional[bool],
        extract_images: Optional[bool],
    ) -> dict[str, Union[str, bool]]:
        params: dict[str, Union[str, bool]] = {
            "url": url,
        }
        if include_raw_html is not None:
            params["includeRawHtml"] = include_raw_html
        if render_js is not None:
            params["renderJs"] = render_js
        if extract_images is not None:
            params["extractImages"] = extract_images
        return params

    def _parse_search_response(
        self,
        response: httpx.Response,
        output_type: Literal["searchResults", "sourcedAnswer", "structured"],
        structured_output_schema: Union[type[BaseModel], str, None],
        include_sources: Optional[bool],
    ) -> Any:
        response_data: Any = response.json()
        if output_type == "searchResults":
            return LinkupSearchResults.model_validate(response_data)
        elif output_type == "sourcedAnswer":
            return LinkupSourcedAnswer.model_validate(response_data)
        elif output_type == "structured":
            if structured_output_schema is None:
                raise ValueError(
                    "structured_output_schema must be provided when output_type is 'structured'"
                )
            # HACK: we assume that `include_sources` will default to False, since the API output can
            # be arbitrary so we can't guess if it includes sources or not
            if include_sources:
                if not isinstance(structured_output_schema, str) and issubclass(
                    structured_output_schema, BaseModel
                ):
                    response_data["data"] = structured_output_schema.model_validate(
                        response_data["data"]
                    )
                return LinkupSearchStructuredResponse.model_validate(response_data)
            if not isinstance(structured_output_schema, str) and issubclass(
                structured_output_schema, BaseModel
            ):
                return structured_output_schema.model_validate(response_data)
            return response_data
        else:
            raise ValueError(f"Unexpected output_type value: '{output_type}'")

    def _parse_fetch_response(self, response: httpx.Response) -> LinkupFetchResponse:
        return LinkupFetchResponse.model_validate(response.json())



"""Example of a structured search.

With `output_type` set to `structured`, the Linkup search can be used to require any arbitrary data
structure, based on a JSON schema or a `pydantic.BaseModel`. This can be used with a well defined
and documented schema to steer the Linkup search in any direction.

To use this script, copy the `.env.example` file at the root of the repository inside a `.env`, and
fill the missing values, or pass a Linkup API key to the `LinkupClient` initialization.
"""

from dotenv import load_dotenv
from pydantic import BaseModel, Field
from rich import print

from linkup import LinkupClient


class Event(BaseModel):
    date: str = Field(description="The date of the event")
    description: str = Field(description="The description of the event")


class Events(BaseModel):
    events: list[Event] = Field(description="The list of events")


load_dotenv()
client = LinkupClient()

response = client.search(
    query="What are the 3 major events in the life of Abraham Lincoln?",
    depth="standard",  # or "deep"
    output_type="structured",
    structured_output_schema=Events,  # or json.dumps(Events.model_json_schema())
    include_sources=False,
)
print(response)



from ._version import __version__
from .client import LinkupClient
from .errors import (
    LinkupAuthenticationError,
    LinkupFailedFetchError,
    LinkupInsufficientCreditError,
    LinkupInvalidRequestError,
    LinkupNoResultError,
    LinkupTooManyRequestsError,
    LinkupUnknownError,
)
from .types import (
    LinkupFetchResponse,
    LinkupSearchImageResult,
    LinkupSearchResults,
    LinkupSearchStructuredResponse,
    LinkupSearchTextResult,
    LinkupSource,
    LinkupSourcedAnswer,
)

__all__ = [
    "__version__",
    "LinkupClient",
    "LinkupAuthenticationError",
    "LinkupFailedFetchError",
    "LinkupInsufficientCreditError",
    "LinkupInvalidRequestError",
    "LinkupNoResultError",
    "LinkupTooManyRequestsError",
    "LinkupUnknownError",
    "LinkupFetchResponse",
    "LinkupSearchImageResult",
    "LinkupSearchResults",
    "LinkupSearchStructuredResponse",
    "LinkupSearchTextResult",
    "LinkupSource",
    "LinkupSourcedAnswer",
]



import json
from datetime import date
from typing import Any

import pytest
from httpx import Response
from pydantic import BaseModel
from pytest_mock import MockerFixture

from linkup import (
    LinkupAuthenticationError,
    LinkupClient,
    LinkupInvalidRequestError,
    LinkupSearchResults,
    LinkupSource,
    LinkupSourcedAnswer,
    LinkupUnknownError,
)
from linkup.errors import (
    LinkupFailedFetchError,
    LinkupInsufficientCreditError,
    LinkupNoResultError,
    LinkupTooManyRequestsError,
)
from linkup.types import (
    LinkupFetchResponse,
    LinkupSearchImageResult,
    LinkupSearchStructuredResponse,
    LinkupSearchTextResult,
)


class Company(BaseModel):
    name: str
    creation_date: str
    website_url: str
    founders_names: list[str]


test_search_parameters = [
    (
        {"query": "query", "depth": "standard", "output_type": "searchResults"},
        {"q": "query", "depth": "standard", "outputType": "searchResults"},
        b"""
        {
            "results": [
                {
                    "type": "text",
                    "name": "foo",
                    "url": "https://foo.com",
                    "content": "lorem ipsum dolor sit amet"
                },
                {"type": "image", "name": "bar", "url": "https://bar.com"}
            ]
        }
        """,
        LinkupSearchResults(
            results=[
                LinkupSearchTextResult(
                    type="text",
                    name="foo",
                    url="https://foo.com",
                    content="lorem ipsum dolor sit amet",
                ),
                LinkupSearchImageResult(
                    type="image",
                    name="bar",
                    url="https://bar.com",
                ),
            ]
        ),
    ),
    (
        {
            "query": "A long query.",
            "depth": "deep",
            "output_type": "searchResults",
            "include_images": True,
            "from_date": date(2023, 1, 1),
            "to_date": date(2023, 12, 31),
            "exclude_domains": ["excluded.com"],
            "include_domains": ["example.com", "example.org"],
            "include_inline_citations": True,
        },
        {
            "q": "A long query.",
            "depth": "deep",
            "outputType": "searchResults",
            "includeImages": True,
            "fromDate": "2023-01-01",
            "toDate": "2023-12-31",
            "excludeDomains": ["excluded.com"],
            "includeDomains": ["example.com", "example.org"],
            "includeInlineCitations": True,
        },
        b'{"results": []}',
        LinkupSearchResults(results=[]),
    ),
    (
        {"query": "query", "depth": "standard", "output_type": "sourcedAnswer"},
        {"q": "query", "depth": "standard", "outputType": "sourcedAnswer"},
        b"""
        {
            "answer": "foo bar baz",
            "sources": [
                {"name": "foo", "url": "https://foo.com", "snippet": "lorem ipsum dolor sit amet"},
                {"name": "bar", "url": "https://bar.com", "snippet": "consectetur adipiscing elit"},
                {"name": "baz", "url": "https://baz.com"}
            ]
        }
        """,
        LinkupSourcedAnswer(
            answer="foo bar baz",
            sources=[
                LinkupSource(
                    name="foo",
                    url="https://foo.com",
                    snippet="lorem ipsum dolor sit amet",
                ),
                LinkupSource(
                    name="bar",
                    url="https://bar.com",
                    snippet="consectetur adipiscing elit",
                ),
                LinkupSource(
                    name="baz",
                    url="https://baz.com",
                    snippet="",
                ),
            ],
        ),
    ),
    (
        {
            "query": "query",
            "depth": "standard",
            "output_type": "structured",
            "structured_output_schema": Company,
        },
        {
            "q": "query",
            "depth": "standard",
            "outputType": "structured",
            "structuredOutputSchema": json.dumps(Company.model_json_schema()),
        },
        b"""
        {
            "name": "Linkup",
            "founders_names": ["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
            "creation_date": "2024",
            "website_url": "https://www.linkup.so/"
        }
        """,
        Company(
            name="Linkup",
            founders_names=["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
            creation_date="2024",
            website_url="https://www.linkup.so/",
        ),
    ),
    (
        {
            "query": "query",
            "depth": "standard",
            "output_type": "structured",
            "structured_output_schema": json.dumps(Company.model_json_schema()),
        },
        {
            "q": "query",
            "depth": "standard",
            "outputType": "structured",
            "structuredOutputSchema": json.dumps(Company.model_json_schema()),
        },
        b"""
        {
            "name": "Linkup",
            "founders_names": ["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
            "creation_date": "2024",
            "website_url": "https://www.linkup.so/"
        }
        """,
        dict(
            name="Linkup",
            founders_names=["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
            creation_date="2024",
            website_url="https://www.linkup.so/",
        ),
    ),
    (
        {
            "query": "query",
            "depth": "standard",
            "output_type": "structured",
            "structured_output_schema": Company,
            "include_sources": True,
        },
        {
            "q": "query",
            "depth": "standard",
            "outputType": "structured",
            "structuredOutputSchema": json.dumps(Company.model_json_schema()),
            "includeSources": True,
        },
        b"""
        {
            "data": {
                "name": "Linkup",
                "founders_names": ["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
                "creation_date": "2024",
                "website_url": "https://www.linkup.so/"
            },
            "sources": [
                {
                    "type": "text",
                    "name": "foo",
                    "url": "https://foo.com",
                    "content": "lorem ipsum dolor sit amet"
                },
                {"type": "image", "name": "bar", "url": "https://bar.com"}
            ]
        }
        """,
        LinkupSearchStructuredResponse(
            data=Company(
                name="Linkup",
                founders_names=["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
                creation_date="2024",
                website_url="https://www.linkup.so/",
            ),
            sources=[
                LinkupSearchTextResult(
                    type="text",
                    name="foo",
                    url="https://foo.com",
                    content="lorem ipsum dolor sit amet",
                ),
                LinkupSearchImageResult(type="image", name="bar", url="https://bar.com"),
            ],
        ),
    ),
    (
        {
            "query": "query",
            "depth": "standard",
            "output_type": "structured",
            "structured_output_schema": json.dumps(Company.model_json_schema()),
            "include_sources": True,
        },
        {
            "q": "query",
            "depth": "standard",
            "outputType": "structured",
            "structuredOutputSchema": json.dumps(Company.model_json_schema()),
            "includeSources": True,
        },
        b"""
        {
            "data": {
                "name": "Linkup",
                "founders_names": ["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
                "creation_date": "2024",
                "website_url": "https://www.linkup.so/"
            },
            "sources": [
                {
                    "type": "text",
                    "name": "foo",
                    "url": "https://foo.com",
                    "content": "lorem ipsum dolor sit amet"
                },
                {"type": "image", "name": "bar", "url": "https://bar.com"}
            ]
        }
        """,
        LinkupSearchStructuredResponse(
            data=dict(
                name="Linkup",
                founders_names=["Philippe Mizrahi", "Denis Charrier", "Boris Toledano"],
                creation_date="2024",
                website_url="https://www.linkup.so/",
            ),
            sources=[
                LinkupSearchTextResult(
                    type="text",
                    name="foo",
                    url="https://foo.com",
                    content="lorem ipsum dolor sit amet",
                ),
                LinkupSearchImageResult(type="image", name="bar", url="https://bar.com"),
            ],
        ),
    ),
]


@pytest.mark.parametrize(
    "search_kwargs, expected_request_params, mock_request_response_content, "
    "expected_search_response",
    test_search_parameters,
)
def test_search(
    mocker: MockerFixture,
    client: LinkupClient,
    search_kwargs: dict[str, Any],
    expected_request_params: dict[str, Any],
    mock_request_response_content: bytes,
    expected_search_response: Any,
) -> None:
    mocker.patch("linkup.client.date").today.return_value = date(2000, 1, 1)
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._request",
        return_value=Response(
            status_code=200,
            content=mock_request_response_content,
        ),
    )

    search_response: Any = client.search(**search_kwargs)
    request_mock.assert_called_once_with(
        method="POST",
        url="/search",
        json=expected_request_params,
        timeout=None,
    )
    assert search_response == expected_search_response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "search_kwargs, expected_request_params, mock_request_response_content, "
    "expected_search_response",
    test_search_parameters,
)
async def test_async_search(
    mocker: MockerFixture,
    client: LinkupClient,
    search_kwargs: dict[str, Any],
    expected_request_params: dict[str, Any],
    mock_request_response_content: bytes,
    expected_search_response: Any,
) -> None:
    mocker.patch("linkup.client.date").today.return_value = date(2000, 1, 1)
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._async_request",
        return_value=Response(
            status_code=200,
            content=mock_request_response_content,
        ),
    )

    search_response: Any = await client.async_search(**search_kwargs)
    request_mock.assert_called_once_with(
        method="POST",
        url="/search",
        json=expected_request_params,
        timeout=None,
    )
    assert search_response == expected_search_response


test_search_error_parameters = [
    (
        403,
        b"""
        {
            "error": {
                "code": "FORBIDDEN",
                "message": "Forbidden action",
                "details": []
            }
        }
        """,
        LinkupAuthenticationError,
    ),
    (
        401,
        b"""
        {
            "error": {
                "code": "UNAUTHORIZED",
                "message": "Unauthorized action",
                "details": []
            }
        }
        """,
        LinkupAuthenticationError,
    ),
    (
        429,
        b"""
        {
            "error": {
                "code": "INSUFFICIENT_FUNDS_CREDITS",
                "message": "You do not have enough credits to perform this request.",
                "details": []
            }
        }
        """,
        LinkupInsufficientCreditError,
    ),
    (
        429,
        b"""
        {
            "error": {
                "code": "TOO_MANY_REQUESTS",
                "message": "Too many requests.",
                "details": []
            }
        }
        """,
        LinkupTooManyRequestsError,
    ),
    (
        429,
        b"""
        {
            "error": {
                "code": "FOOBAR",
                "message": "Foobar",
                "details": []
            }
        }
        """,
        LinkupUnknownError,
    ),
    (
        400,
        b"""
        {
            "error": {
                "code": "VALIDATION_ERROR",
                "message": "Validation failed",
                "details": [
                    {
                        "field": "structuredOutputSchema",
                        "message": "structuredOutputSchema must be valid JSON schema of type"
                    }
                ]
            }
        }
        """,
        LinkupInvalidRequestError,
    ),
    (
        400,
        b"""
        {
            "error": {
                "code": "SEARCH_QUERY_NO_RESULT",
                "message": "The query did not yield any result",
                "details": []
            }
        }
        """,
        LinkupNoResultError,
    ),
    (
        500,
        b"""
        {
            "error": {
                "code": "INTERNAL_SERVER_ERROR",
                "message": "Internal server error",
                "details": []
            }
        }
        """,
        LinkupUnknownError,
    ),
]


@pytest.mark.parametrize(
    "mock_request_response_status_code, mock_request_response_content, expected_exception",
    test_search_error_parameters,
)
def test_search_error(
    mocker: MockerFixture,
    client: LinkupClient,
    mock_request_response_status_code: int,
    mock_request_response_content: bytes,
    expected_exception: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._request",
        return_value=Response(
            status_code=mock_request_response_status_code,
            content=mock_request_response_content,
        ),
    )

    with pytest.raises(expected_exception):
        client.search(query="query", depth="standard", output_type="searchResults")
    request_mock.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mock_request_response_status_code, mock_request_response_content, expected_exception",
    test_search_error_parameters,
)
async def test_async_search_error(
    mocker: MockerFixture,
    client: LinkupClient,
    mock_request_response_status_code: int,
    mock_request_response_content: bytes,
    expected_exception: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._async_request",
        return_value=Response(
            status_code=mock_request_response_status_code,
            content=mock_request_response_content,
        ),
    )

    with pytest.raises(expected_exception):
        await client.async_search(query="query", depth="standard", output_type="searchResults")
    request_mock.assert_called_once()


test_fetch_parameters = [
    (
        {"url": "https://example.com"},
        {"url": "https://example.com"},
        b'{"markdown": "Some web page content"}',
        LinkupFetchResponse(markdown="Some web page content", raw_html=None),
    ),
    (
        {"url": "https://example.com", "include_raw_html": True, "render_js": True},
        {"url": "https://example.com", "includeRawHtml": True, "renderJs": True},
        b'{"markdown": "#Some web page content", "rawHtml": "<html>...</html>"}',
        LinkupFetchResponse(markdown="#Some web page content", raw_html="<html>...</html>"),
    ),
]


@pytest.mark.parametrize(
    "fetch_kwargs, expected_request_params, mock_request_response_content, expected_fetch_response",
    test_fetch_parameters,
)
def test_fetch(
    mocker: MockerFixture,
    client: LinkupClient,
    fetch_kwargs: dict[str, Any],
    expected_request_params: dict[str, Any],
    mock_request_response_content: bytes,
    expected_fetch_response: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._request",
        return_value=Response(
            status_code=200,
            content=mock_request_response_content,
        ),
    )

    fetch_response: Any = client.fetch(**fetch_kwargs)
    request_mock.assert_called_once_with(
        method="POST",
        url="/fetch",
        json=expected_request_params,
        timeout=None,
    )
    assert fetch_response == expected_fetch_response


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "fetch_kwargs, expected_request_params, mock_request_response_content, expected_fetch_response",
    test_fetch_parameters,
)
async def test_async_fetch(
    mocker: MockerFixture,
    client: LinkupClient,
    fetch_kwargs: dict[str, Any],
    expected_request_params: dict[str, Any],
    mock_request_response_content: bytes,
    expected_fetch_response: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._async_request",
        return_value=Response(
            status_code=200,
            content=mock_request_response_content,
        ),
    )

    fetch_response: Any = await client.async_fetch(**fetch_kwargs)
    request_mock.assert_called_once_with(
        method="POST",
        url="/fetch",
        json=expected_request_params,
        timeout=None,
    )
    assert fetch_response == expected_fetch_response


test_fetch_error_parameters = [
    (
        400,
        b"""
        {
            "error": {
                "code": "FETCH_ERROR",
                "message": "Could not fetch the URL",
                "details": []
            }
        }
        """,
        LinkupFailedFetchError,
    ),
    (
        400,
        b"""
        {
            "error": {
                "code": "VALIDATION_ERROR",
                "message": "Validation failed",
                "details": [
                    {
                        "field": "url",
                        "message": "url must be a valid URL"
                    }
                ]
            }
        }
        """,
        LinkupInvalidRequestError,
    ),
]


@pytest.mark.parametrize(
    "mock_request_response_status_code, mock_request_response_content, expected_exception",
    test_fetch_error_parameters,
)
def test_fetch_error(
    mocker: MockerFixture,
    client: LinkupClient,
    mock_request_response_status_code: int,
    mock_request_response_content: bytes,
    expected_exception: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._request",
        return_value=Response(
            status_code=mock_request_response_status_code,
            content=mock_request_response_content,
        ),
    )

    with pytest.raises(expected_exception):
        client.fetch(url="https://example.com")
    request_mock.assert_called_once()


@pytest.mark.asyncio
@pytest.mark.parametrize(
    "mock_request_response_status_code, mock_request_response_content, expected_exception",
    test_fetch_error_parameters,
)
async def test_async_fetch_error(
    mocker: MockerFixture,
    client: LinkupClient,
    mock_request_response_status_code: int,
    mock_request_response_content: bytes,
    expected_exception: Any,
) -> None:
    request_mock = mocker.patch(
        "linkup.client.LinkupClient._async_request",
        return_value=Response(
            status_code=mock_request_response_status_code,
            content=mock_request_response_content,
        ),
    )

    with pytest.raises(expected_exception):
        await client.async_fetch(url="https://example.com")
    request_mock.assert_called_once()



"""Example of search results search.

The Linkup search can output raw search results which can then be re-used in different use-cases,
for instance in a RAG system, with the `output_type` parameter set to `searchResults`.

To use this script, copy the `.env.example` file at the root of the repository inside a `.env`, and
fill the missing values, or pass a Linkup API key to the `LinkupClient` initialization.
"""

from dotenv import load_dotenv
from rich import print

from linkup import LinkupClient

load_dotenv()
client = LinkupClient()

response = client.search(
    query="What are the 3 major events in the life of Abraham Lincoln?",
    depth="standard",  # or "deep"
    output_type="searchResults",
)
print(response)



"""Example of web page fetch.

The Linkup fetch can output the content of a web page as a cleaned up markdown.

To use this script, copy the `.env.example` file at the root of the repository inside a `.env`, and
fill the missing values, or pass a Linkup API key to the `LinkupClient` initialization.
"""

from dotenv import load_dotenv
from rich import print

from linkup import LinkupClient

load_dotenv()
client = LinkupClient()

response = client.fetch(
    url="https://docs.linkup.so",
)
print(response)



from typing import Any, Literal, Optional, Union

from pydantic import BaseModel, ConfigDict, Field


class LinkupSearchTextResult(BaseModel):
    """A text result from a Linkup search.

    Attributes:
        type: The type of the search result, in this case "text".
        name: The name of the search result.
        url: The URL of the search result.
        content: The text of the search result.
    """

    type: Literal["text"]
    name: str
    url: str
    content: str


class LinkupSearchImageResult(BaseModel):
    """An image result from a Linkup search.

    Attributes:
        type: The type of the search result, in this case "image".
        name: The name of the image result.
        url: The URL of the image result.
    """

    type: Literal["image"]
    name: str
    url: str


class LinkupSearchResults(BaseModel):
    """The results of the Linkup search.

    Attributes:
        results: The results of the Linkup search.
    """

    results: list[Union[LinkupSearchTextResult, LinkupSearchImageResult]]


class LinkupSource(BaseModel):
    """A source supporting a Linkup answer.

    Attributes:
        name: The name of the source.
        url: The URL of the source.
        snippet: The text excerpt supporting the Linkup answer. Can be empty for image sources.
    """

    name: str
    url: str
    snippet: str = ""


class LinkupSourcedAnswer(BaseModel):
    """A Linkup answer, with the sources supporting it.

    Attributes:
        answer: The answer text.
        sources: The sources supporting the answer.
    """

    answer: str
    sources: list[LinkupSource]


class LinkupSearchStructuredResponse(BaseModel):
    """A Linkup `search` structured response, with the sources supporting it.

    Attributes:
        data: The answer data, either as a Pydantic model or an arbitrary JSON structure.
        sources: The sources supporting the answer.
    """

    data: Any
    sources: list[Union[LinkupSearchTextResult, LinkupSearchImageResult]]


class LinkupFetchImageExtraction(BaseModel):
    """An image extraction from a Linkup web page fetch.

    Attributes:
        alt: The alt text of the image.
        url: The URL of the image.
    """

    alt: str
    url: str


class LinkupFetchResponse(BaseModel):
    """The response from a Linkup web page fetch.

    Attributes:
        markdown: The cleaned up markdown content.
        raw_html: The optional raw HTML content.
        images: The optional list of image URLs.
    """

    model_config = ConfigDict(populate_by_name=True)

    markdown: str
    raw_html: Optional[str] = Field(default=None, validation_alias="rawHtml")
    images: Optional[list[LinkupFetchImageExtraction]] = Field(default=None)



class LinkupInvalidRequestError(Exception):
    """Invalid request error, raised when the Linkup API returns a 400 status code.

    It is returned by the Linkup API when the request is invalid, typically when a mandatory
    parameter is missing, or isn't valid (type, structure, etc.).
    """

    pass


class LinkupNoResultError(Exception):
    """No result error, raised when the Linkup API returns a 400 status code.

    It is returned by the Linkup API when the search query did not yield any result.
    """

    pass


class LinkupAuthenticationError(Exception):
    """Authentication error, raised when the Linkup API returns a 403 status code.

    It is returned when there is an authentication issue, typically when the API key is not valid.
    """

    pass


class LinkupInsufficientCreditError(Exception):
    """Insufficient credit error, raised when the Linkup API returns a 429 status code.

    It is returned when you have run out of credits.
    """

    pass


class LinkupTooManyRequestsError(Exception):
    """Too many requests error, raised when the Linkup API returns a 429 status code.

    It is returned when you are sending too many requests at a time.
    """

    pass


class LinkupFailedFetchError(Exception):
    """Failed fetch error, raised when the Linkup API search returns a 400 status code.

    It is returned when the Linkup API failed to fetch the content of an URL due to technical
    reasons.
    """

    pass


class LinkupUnknownError(Exception):
    """Unknown error, raised when the Linkup API returns an unknown status code."""

    pass



"""Example of asynchronous search.

All Linkup entrypoints come with an asynchronous version. This snippet demonstrates how to run
multiple asynchronous searches concurrently, which decreases by a lot the total computation
duration.

To use this script, copy the `.env.example` file at the root of the repository inside a `.env`, and
fill the missing values, or pass a Linkup API key to the `LinkupClient` initialization.
"""

import asyncio
import time

from dotenv import load_dotenv
from rich import print

from linkup import LinkupClient

load_dotenv()
client = LinkupClient()

queries: list[str] = [
    "What are the 3 major events in the life of Abraham Lincoln?",
    "What are the 3 major events in the life of George Washington?",
]

t0: float = time.time()


async def search(idx: int, query: str) -> None:
    """Run an asynchronous search and display its results and the duration from the beginning."""
    response = await client.async_search(
        query=query,
        depth="standard",  # or "deep"
        output_type="searchResults",  # or "sourcedAnswer" or "structured"
    )
    print(f"{idx + 1}: {time.time() - t0:.3f}s")
    print(response)
    print("-" * 100)


async def main() -> None:
    """Run multiple asynchronous searches concurrently."""
    coroutines = [search(idx=idx, query=query) for idx, query in enumerate(queries)]
    await asyncio.gather(*coroutines)
    print(f"Total time: {time.time() - t0:.3f}s")


asyncio.run(main())
